#include "precompile.h"

#include <f8/udplog.h>

#include "gridservice.h"
#include "human.h"
#include "room.h"
#include "gridcell.h"
#include "bullet.h"

/*
  1 2 3
  4 5 6
  7 8 9
 */

GridService::GridService()
{

}

GridService::~GridService()
{

}

void GridService::Init(int width, int height, int cell_width)
{
    map_width_ = width;
    map_height_ = height;
    cell_width_ = cell_width;

    cell_count_per_row_ = (map_width_ / cell_width + 1) + 2;
    cell_count_per_col_ = (map_height_ / cell_width + 1) + 2;
    max_grid_id_ = cell_count_per_row_ * cell_count_per_col_;
    cells_.reserve(max_grid_id_ + 1);
    for (int i = 0; i < max_grid_id_ + 1; ++i) {
        cells_.push_back(std::make_shared<GridCell>());
    }

    grid_offset_arr_[0] = -(cell_count_per_row_ - 1);
    grid_offset_arr_[1] = -cell_count_per_row_;
    grid_offset_arr_[2] = -(cell_count_per_row_ + 1);

    grid_offset_arr_[3] = -1;
    grid_offset_arr_[4] = 0;
    grid_offset_arr_[5] = 1;

    grid_offset_arr_[6] = cell_count_per_row_ - 1;
    grid_offset_arr_[7] = cell_count_per_row_;
    grid_offset_arr_[8] = cell_count_per_row_ + 1;

    min_x_ = cell_width_ * 1;
    min_y_ = cell_width_ * 1;
    max_x_ = cell_width_ * (cell_count_per_row_ - 1);
    max_y_ = cell_width_ * (cell_count_per_col_ - 1);
}

void GridService::UnInit()
{
}

bool GridService::BroderOverFlow(int x, int y)
{
    return x < min_x_ || x > max_x_ || y < min_y_ || y > max_y_;
}

void GridService::GetAllCells(Room* room, int grid_id, std::set<GridCell*>& grid_list)
{
    for (size_t i = 1; i <= 9; ++i) {
        GetGridList(grid_id, i, grid_list);
    }
}

void GridService::GetAllCellsByXy(Room* room, int x, int y, std::set<GridCell*>& grid_list)
{
    int new_x = x + cell_width_;
    int new_y = y + cell_width_;
    int new_grid_id = new_x/cell_width_ + (new_y/cell_width_) * cell_count_per_row_;
    if (BroderOverFlow(new_x, new_y)) {
        #if 1
        return;
        #else
        A8_ABORT();
        #endif
    }
    GetAllCells(room, new_grid_id, grid_list);
}

bool GridService::CanAdd(float x, float y)
{
    if (x < 0.001f) {
        return false;
    }
    if (y < 0.00001f) {
        return false;
    }
    if (x > map_width_) {
        return false;
    }
    if (y > map_height_) {
        return false;
    }
    int new_x = (int)x + cell_width_;
    int new_y = (int)y + cell_width_;
    return !BroderOverFlow(new_x, new_y);
}

void GridService::ClearRoomData(Room* room)
{
    if (room->GetRoomIdx() <= 0) {
        A8_ABORT();
    }
    for (int i = 0; i < (max_grid_id_ + 1); ++i) {
        auto cell = cells_.at(i);
        cell->ClearRoomData(room);
    }
}

void GridService::AddCreature(Creature* c)
{
    int x = (int)c->GetPos().GetX() + cell_width_;
    int y = (int)c->GetPos().GetZ() + cell_width_;
    if (BroderOverFlow(x, y)) {
#if 1
        f8::UdpLog::Instance()->Warning
            (
             "AddCreature broder over pos:%f,%f,%f new:%f,%f",
             {
                 c->GetPos().GetX(),
                     c->GetPos().GetY(),
                     c->GetPos().GetZ(),
                     x,
                     y
             }
             );
        return;
#else
        A8_ABORT();
#endif
    }
    c->SetGridId(x/cell_width_ + (y/cell_width_) * cell_count_per_row_);
    if (c->GetGridId() == 0 || c->GetGridId() > max_grid_id_) {
        A8_ABORT();
    }
    cells_.at(c->GetGridId())->AddCreature(c);
    GetAllCells(c->room, c->GetGridId(), c->GetGridList());
}

void GridService::RemoveCreature(Creature* c)
{
    auto cell = cells_.at(c->GetGridId());
    cell->RemoveCreature(c);
}

void GridService::MoveCreature(Creature* c)
{
    int new_x = (int)c->GetPos().GetX() + cell_width_;
    int new_y = (int)c->GetPos().GetZ() + cell_width_;
    int new_grid_id = new_x/cell_width_ + (new_y/cell_width_) * cell_count_per_row_;
    if (BroderOverFlow(new_x, new_y)) {
#if 1
        f8::UdpLog::Instance()->Warning
            (
             "MoveCreature broder over pos:%f,%f,%f new:%f,%f",
             {
                 c->GetPos().GetX(),
                     c->GetPos().GetY(),
                     c->GetPos().GetZ(),
                     new_x,
                     new_y
             }
             );
        return;
#else
        A8_ABORT();
#endif
    }
    if (new_grid_id == 0 || new_grid_id > max_grid_id_) {
        A8_ABORT();
    }
    if (new_grid_id != c->GetGridId()) {
        std::set<GridCell*> inc_grid_list;
        std::set<GridCell*> dec_grid_list;
        std::set<GridCell*> old_grid_list = c->GetGridList();
        ComputeDiff(c->room,
                    c->GetGridId(),
                    new_grid_id,
                    c->GetGridList(),
                    inc_grid_list,
                    dec_grid_list);
        cells_.at(c->GetGridId())->RemoveCreature(c);
        cells_.at(new_grid_id)->AddCreature(c);
        c->SetGridId(new_grid_id);
        c->OnGridListChange(old_grid_list, inc_grid_list, dec_grid_list);
    }
}

void GridService::AddRoomEntity(Room* room, Entity* entity)
{
    assert(!entity->IsEntityType(ET_Player));
    int x = (int)entity->GetPos().GetX() + cell_width_;
    int y = (int)entity->GetPos().GetZ() + cell_width_;
    if (BroderOverFlow(x, y)) {
#if 1
        f8::UdpLog::Instance()->Warning
            (
             "AddRoomEntity broder over pos:%f,%f,%f new:%f,%f",
             {
                 entity->GetPos().GetX(),
                     entity->GetPos().GetY(),
                     entity->GetPos().GetZ(),
                     x,
                     y
             }
             );
        return;
#else
        A8_ABORT();
#endif
    }
    entity->SetGridId(x/cell_width_ + (y/cell_width_) * cell_count_per_row_);
    if (entity->GetGridId() == 0 || entity->GetGridId() > max_grid_id_) {
        A8_ABORT();
    }
    cells_.at(entity->GetGridId())->AddRoomEntity(room, entity);
}

void GridService::DelRoomEntity(Room* room, Entity* entity)
{
    auto cell = cells_.at(entity->GetGridId());
    cell->RemoveRoomEntity(room, entity);
}

bool GridService::CreatureInGridList(Creature* c, std::set<GridCell*>& grid_list)
{
    for (auto& cell : grid_list) {
        if (cell->CreatureExists(c)) {
            return true;
        }
    }
    return false;
}

bool GridService::EntityInGridList(Room* room, Entity* entity, std::set<GridCell*>& grid_list)
{
    for (auto& cell : grid_list) {
        if (cell->EntityExists(room, entity)) {
            return true;
        }
    }
    return false;
}

bool GridService::InView(int a_grid, int b_grid)
{
    return a_grid + grid_offset_arr_[0] == b_grid ||
        a_grid + grid_offset_arr_[1] == b_grid ||
        a_grid + grid_offset_arr_[2] == b_grid ||
        a_grid + grid_offset_arr_[3] == b_grid ||
        a_grid + grid_offset_arr_[4] == b_grid ||
        a_grid + grid_offset_arr_[5] == b_grid ||
        a_grid + grid_offset_arr_[6] == b_grid ||
        a_grid + grid_offset_arr_[7] == b_grid ||
        a_grid + grid_offset_arr_[8] == b_grid;
}

bool GridService::InView(int grid_id, float x, float y)
{
    int b_grid_id = x/cell_width_ + (y/cell_width_) * cell_count_per_row_;
    bool in_view = InView(grid_id, b_grid_id);
#ifdef MYDEBUG
    if (in_view) {
        f8::UdpLog::Instance()->Debug("InView %d %d %f %f",
                                      {
                                       grid_id,
                                       b_grid_id,
                                       x,
                                       y
                                      });
    }
#endif
    return in_view;
}

void GridService::GetGridList(int grid_id, int offset,
                       std::set<GridCell*>& grid_list)
{
    if (offset <1 || offset > 9) {
        A8_ABORT();
    }
    if (grid_id <= 0 || grid_id >= max_grid_id_) {
        return;
    }
    int tmp_grid_id = grid_id + grid_offset_arr_[offset - 1];
    if (tmp_grid_id > 0 && tmp_grid_id < max_grid_id_) {
        grid_list.insert(cells_.at(tmp_grid_id).get());
    }
}

void GridService::ComputeDiff(Room* room,
                              int old_grid_id,
                              int new_grid_id,
                              std::set<GridCell*>& grid_list,
                              std::set<GridCell*>& inc_grid_list,
                              std::set<GridCell*>& dec_grid_list)
{
    #if 1
    {
        std::set<GridCell*> new_grid_list;
        GetAllCells(room, new_grid_id, new_grid_list);
        for (auto& cell : new_grid_list) {
            if (grid_list.find(cell) == grid_list.end()) {
                inc_grid_list.insert(cell);
            }
        }
        for (auto& cell : grid_list) {
            if (new_grid_list.find(cell) == new_grid_list.end()) {
                dec_grid_list.insert(cell);
            }
        }
    }
    #endif

    grid_list.clear();
    GetAllCells(room, new_grid_id, grid_list);
}

void GridService::TraverseLayer0EntityList(std::set<GridCell*>& grid_list,
                                        std::function<void (Entity*, bool&)> func)
{
    bool stop = false;
    for (auto& cell : grid_list) {
        cell->TraverseLayer0EntityList(func, stop);
        if (stop) {
            return;
        }
    }
}

void GridService::TraverseLayer1EntityList(int room_idx,
                                        std::set<GridCell*>& grid_list,
                                        std::function<void (Entity*, bool&)> func)
{
    bool stop = false;
    for (auto& cell : grid_list) {
        cell->TraverseLayer1EntityList(func, room_idx, stop);
        if (stop) {
            return;
        }
    }
}

void GridService::TraverseAllLayerEntityList(int room_idx,
                                          std::set<GridCell*>& grid_list,
                                          std::function<void (Entity*, bool&)> func)
{
    bool stop = false;
    for (auto& cell : grid_list) {
        cell->TraverseAllLayerEntityList(func, room_idx, stop);
        if (stop) {
            return;
        }
    }
}

void GridService::TraverseAllLayerHumanList(int room_idx,
                                         std::set<GridCell*>& grid_list,
                                         std::function<void (Human*, bool&)> func)
{
    bool stop = false;
    for (auto& cell : grid_list) {
        cell->TraverseHumanList(func, room_idx, stop);
        if (stop) {
            return;
        }
    }
}

void GridService::TraverseCreatures(int room_idx,
                                 std::set<GridCell*>& grid_list,
                                 std::function<void (Creature*, bool&)> func)
{
    bool stop = false;
    for (auto& cell : grid_list) {
        cell->TraverseCreatures(func, room_idx, stop);
        if (stop) {
            return;
        }
    }
}

void GridService::TraverseObstacles(int room_idx,
                                    std::set<GridCell*>& grid_list,
                                    std::function<void (Obstacle*, bool&)> func)
{
    bool stop = false;
    for (auto& cell : grid_list) {
        cell->TraverseObstacles(func, room_idx, stop);
        if (stop) {
            return;
        }
    }
}

void GridService::DeatchHuman(Human* target)
{
    #if 1
    target->room->TraverseHumanList
        (
         [target] (Human* hum)
         {
             hum->AddOutObjects(target);
             return true;
         });
    for (auto& cell : target->GetGridList()) {
        cell->RemoveCreature(target);
    }
    #else
    for (auto& cell : target->GetGridList()) {
        bool stop = false;
        cell->TraverseHumanList
            (
             [target] (Human* hum, bool& stop)
             {
                 hum->AddOutObjects(target);
             },
             target->room->GetRoomIdx(),
             stop);
        cell->RemoveCreature(target);
    }
    #endif
    target->SetGridId(0);
    target->GetGridList().clear();
}

void GridService::MoveBullet(Bullet* bullet)
{
    int new_x = (int)bullet->GetPos().GetX() + cell_width_;
    int new_y = (int)bullet->GetPos().GetZ() + cell_width_;
    int new_grid_id = new_x/cell_width_ + (new_y/cell_width_) * cell_count_per_row_;
    if (BroderOverFlow(new_x, new_y)) {
#if 1
        f8::UdpLog::Instance()->Warning
            (
             "MoveBullet broder over pos:%f,%f,%f new:%f,%f",
             {
                 bullet->GetPos().GetX(),
                     bullet->GetPos().GetY(),
                     bullet->GetPos().GetZ(),
                     new_x,
                     new_y
             }
             );
        return;
#else
        A8_ABORT();
#endif
    }
    if (new_grid_id == 0 || new_grid_id > max_grid_id_) {
        A8_ABORT();
    }
    if (new_grid_id != bullet->GetGridId()) {
        std::set<GridCell*> inc_grid_list;
        std::set<GridCell*> dec_grid_list;
        std::set<GridCell*> old_grid_list = bullet->GetGridList();
        ComputeDiff(bullet->room,
                    bullet->GetGridId(),
                    new_grid_id,
                    bullet->GetGridList(),
                    inc_grid_list,
                    dec_grid_list);
        bullet->SetGridId(new_grid_id);
    }
}
